# Internal Library
import aepp
from aepp import config
from copy import deepcopy
import typing
from concurrent import futures as _futures


class Profile:

    def __init__(self, **kwargs):
        self.header = deepcopy(aepp.config._header)
        self.header['Accept'] = "application/vnd.adobe.xdm+json"
        self.header.update(**kwargs)
        # same endpoint than segmentation
        self.endpoint = config._endpoint+config._endpoint_segmentation

    def getEntity(self, schema_name: str = "_xdm.context.profile", entityId: str = None, entityIdNS: str = None, mergePoliciyId: str = None, **kwargs):
        """
        Returns an entity by ID or Namespace.
        Arguments:
            schema_name : REQUIRED : class name of the schema to be retrieved. default : _xdm.context.profile
            entityId : OPTIONAL : idenitity ID
            entityIdNS : OPTIONAL : Identity Namespace code. Required if entityId is used.
            mergePoliciyId : OPTIONAL : Id of the merge policy.
        Possible kwargs:
            fields : path of the elements to be retrieved, separated by comma. Ex : "person.name.firstName,person.name.lastName"
            relatedSchema_name : If schema.name is "_xdm.context.experienceevent", this value must specify the schema for the profile entity that the time series events are related to. 
            relatedEntityId : ID of the entity that the ExperienceEvents are associated with. Used when looking up ExperienceEvents. For Native XID lookup, use relatedEntityId=<XID> and leave relatedEntityIdNS absent; 
            For ID:NS lookup, use both relatedEntityId and relatedEntityIdNS fields.
            relatedEntityIdNS : Identity Namespace code of the related entity ID of ExperienceEvent. Used when looking up ExperienceEvents. If this field is used, entityId cannot be empty.
            startTime : Start time of Time range filter for ExperienceEvents. Should be at millisecond granularity. Included. Default: From beginning.
            endTime : End time of Time range filter for ExperienceEvents. Should be at millisecond granularity. Excluded. Default: To the end.
            limit : Number of records to return from the result. Only for time-series objects. Default: 1000
        """
        path = "/access/entities"
        params = {"schema.name":  schema_name, "entityId": entityId,
                  "entityIdNS": entityIdNS, "mergePoliciyId": mergePoliciyId}
        if schema_name == "_xdm.context.experienceevent":
            params["relatedSchema.name"] = kwargs.get(
                "relatedSchema_name", "_xdm.context.profile")
            params["relatedEntityId"] = kwargs.get(
                "relatedEntityId", entityId)
            params["relatedEntityIdNS"] = kwargs.get(
                "relatedEntityIdNS", None)
            params["limit"] = kwargs.get("limit", 1000)
            params["startTime"] = kwargs.get("startTime", None)
            params["endTime"] = kwargs.get("endTime", None)
        params["fields"] = kwargs.get("fields", None)
        res = aepp._getData(self.endpoint+path,
                            params=params, headers=self.header)
        return res

    def getEntities(self, request_data: dict = None):
        """
        Get a number of different identities from ID or namespaces.
        Argument:
            request_data : Required : A dictionary that contains fields for the search.
            Example
            {
                "schema": {
                    "name": "_xdm.context.profile"
                },
                "relatedSchema": {
                    "name": "_xdm.context.profile"
                },
                "fields": [
                    "person.name.firstName"
                ],
                "identities": [
                    {
                    "entityId": "69935279872410346619186588147492736556",
                    "entityIdNS": {
                        "code": "CRMID"
                    }
                    }
                ],
                "timeFilter": {
                    "startTime": 1539838505,
                    "endTime": 1539838510
                },
                "limit": 10
                }
        """
        path = "/access/entities"
        if request_data is None or type(request_data) != dict:
            raise Exception("Expected a dictionary to fetch entities")
        res = aepp._postData(self.endpoint + path,
                             data=request_data, headers=self.header)
        return res

    def deleteEntity(self, schema_name: str = None, entityId: str = None, entityIdNS: str = None):
        """
        Delete a specific entity
        Arguments:
            schema_name : REQUIRED : Name of the associated XDM schema.
            entityId : REQUIRED : entity ID
            entityIdNS : OPTIONAL : entity ID Namespace
        """
        path = "/access/entities"
        params = {}
        if schema_name is None:
            raise Exception("Expected a schema name")
        else:
            params["schema.name"] = schema_name
        if entityId is not None:
            params["entityId"] = entityId
        if entityIdNS is not None:
            params["entityIdNS"] = entityIdNS
        res = aepp._getData(self.endpoint+path,
                            params=params, headers=self.header)

    def getMergePolicies(self, limit: int = 100):
        """
        Returns the list of merge policies hosted in this instance.
        Arguments:
            limit : OPTIONAL : amount of merge policies returned by pages.
        """
        path = "/config/mergePolicies"
        params = {"limit": limit}
        res = aepp._getData(self.endpoint+path,
                            params=params, headers=self.header)
        return res

    def getMergePolicy(self, policy_id: str = None):
        """
        Return a specific merge policy.
        Arguments:
            policy_id : REQUIRED : id of the the policy id to be returned.
        """
        if policy_id is None:
            raise Exception("Missing the policy id")
        path = f"/config/mergePolicies/{policy_id}"
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res



import aepp
from aepp import config
from copy import deepcopy
import re
import typing
from dataclasses import dataclass


@dataclass
class _Data:

    def __init__(self):
        self.table_names = {}
        self.schema_ref = {}
        self.ids = {}


class Catalog:
    """
    Catalog class from the AEP API. This class helps you to find where the data are coming from in AEP.
    More details here : https://www.adobe.io/apis/experienceplatform/home/api-reference.html#
    It possess a data attribute that is containing information about your datasets. 
    """

    def __init__(self, **kwargs):
        self.header = deepcopy(aepp.config._header)
        self.header['Accept'] = "application/vnd.adobe.xdm+json"
        self.header.update(**kwargs)
        self.endpoint = config._endpoint+config._endpoint_catalog
        self.data = _Data()

    def getAccounts(self, **kwargs):
        """
        Returns the list of all account.
        Possible kwargs:
            connector : The ID for the Connector this Account params was created from
            description : Filter by user-provided description of the account.
            version : Filter by Semantic version of the account. Updated when the object is modified.
            created : Filter by the Unix timestamp (in milliseconds) when this object was persisted.
            updated : Filter by the Unix timestamp (in milliseconds) for the time of last modification.
            createdUser : Filter by the ID of the user who created this object.
            property : Regex used to filter objects in the response. Ex. property=name~^test.
        # /Accounts/get_accounts
        more details : https://www.adobe.io/apis/experienceplatform/home/api-reference.html
        """
        path = "/accounts/"
        params = {**kwargs}
        res = aepp._getData(self.endpoint+path,
                            headers=self.header, params=params)
        return res

    def getAccount(self, account_id: str = None):
        """
        Get a specific Account based on the id.
        Arguments:
            account_id : REQUIRED : object ID
        """
        if account_id is None:
            raise Exception("Require an account_id")
        path = f"/accounts/{account_id}"
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res

    def createAccount(self, data: dict = None):
        """
        Create a new account.
        Arguments:
            data : REQUIRED : Object that should describe the new connector.
        More information on data requirement can be found here :
        https://www.adobe.io/apis/experienceplatform/home/api-reference.html#/Accounts/post_account
        """
        if type(data) != dict:
            raise TypeError("data should a dictionary")
        elif "params" not in data.keys():
            raise AttributeError("params should be part of the dictionnary")
        path = "/accounts/"
        res = aepp._postData(self.endpoint+path,
                             headers=self.header, data=data)
        return res

    def updateAccount(self, account_id: str = None, data: dict = None):
        """
        Update an account.
        Arguments:
            account_id : REQUIRED : account id to be udpated.
            data : REQUIRED : Object that should describe the new connector.
        """
        if account_id is None:
            raise Exception("Require an account_id")
        elif type(data) != dict:
            raise TypeError("data should a dictionary")
        elif "params" not in data.keys():
            raise AttributeError("params should be part of the dictionnary")
        path = f"/accounts/{account_id}"
        res = aepp._putData(self.endpoint+path,
                            headers=self.header, data=data)
        return res

    def deleteAccount(self, account_id: str = None):
        """
        Delete an account.
        Arguments:
            account_id : REQUIRED : account id to be deleted.
        """
        if account_id is None:
            raise Exception("Require an account_id")
        path = f"/accounts/{account_id}"
        res = aepp._deleteData(self.endpoint+path,
                               headers=self.header)
        return res

    def getBatches(self, **kwargs):
        """
        Retrieve a list of batches.
        Possible kwargs:
            created : Filter by the Unix timestamp (in milliseconds) when this object was persisted.
            createdAfter : Filter by the Unix timestamp (in milliseconds) when this object was persisted.
            limit : Limit response to a specified positive number of objects. Ex. limit=10
            updated : Filter by the Unix timestamp (in milliseconds) for the time of last modification.
            createdUser : Filter by the ID of the user who created this object.
            dataSet : Used to filter on the related object: &dataSet=dataSetId.
            version : Filter by Semantic version of the account. Updated when the object is modified.
            property : Regex used to filter objects in the response. Ex. property=name~^test.
        # /Batches/get_batch
        more details : https://www.adobe.io/apis/experienceplatform/home/api-reference.html
        """
        path = "/batches"
        params = {**kwargs}
        res = aepp._getData(self.endpoint+path,
                            headers=self.header, params=params)
        return res

    def getBatch(self, batch_id: str = None):
        """
        Get a specific batch id.
        Arguments:
            batch_id : REQUIRED : batch ID to be retrieved.
        """
        if batch_id is None:
            raise Exception("batch_id parameter is required.")
        path = f"/batches/{batch_id}"
        res = aepp._getData(self.endpoint+path,
                            headers=self.header)
        return res

    def getConnectors(self, **kwargs):
        """
        Returns the list of connectors
        Possible kwargs:
            name : Filter by the name of this Connector.
            type : Filter by the ingest type for this Connector.
            category : Connectors currently have 2 categories: 1) standard: we are connecting to a known source with known data (ie. other companies data stores) or 2) custom: a generic connector to FTP or S3 etc.
            limit : Limit response to a specified positive number of objects. Ex. limit=10
            version : Filter by Semantic version of the account. Updated when the object is modified.
            property : Regex used to filter objects in the response. Ex. property=name~^test.
        """
        path = "/connectors"
        params = {**kwargs}
        res = aepp._getData(self.endpoint+path,
                            headers=self.header, params=params)
        return res

    def getConnector(self, connector_id: str = None, stats: bool = False):
        """
        Returns a scpecific connector details and its stats if required in the parameter.
        Arguments:
            connector_id : REQUIRED : connector id for the connector to be retrieved.
            stats : OPTIONAL : If set to True, return a 2nd object with the stat of the connector.
        """
        if connector_id is None:
            raise Exception("Expected a connector_id parameter.")
        path = f"/connectors/{connector_id}"
        res = aepp._getData(self.endpoint+path, headers=self.header)
        if stats == False:
            return res
        elif stats:
            stats = self.getConnectorStats(connector_id)
            return res, stats

    def getConnectorStats(self, connector_id: str = None):
        """
        Retrieve the stats for a specific connector.
        Arguments:
            connector_id : REQUIRED : connector id for the connector stats to be retrieved.
        """
        if connector_id is None:
            raise Exception("Expected a connector_id parameter.")
        path = f"/connectors/{connector_id}/stats"
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res

    def getConnections(self, **kwargs):
        """
        Returns list of connections.
        Possible kwargs:
            connector : Filter by the ID for the Connector this Connection was created from.
            parentConnectionId : Used in cases where global/shared data is managed by this connection. The parent connection performs the ETL/Mapping jobs, and this child connection represents a customer’s membership and visibility into the parent.
            name : Filter by the user-facing name of this Connection.
            accountId : Filter by the foreign key to the account where the credentials and related fields of the connector and connection combination is stored.
            description : Filter by the user-provided description of the Connection.
            enabled : Indicates the status of the Connection. Should be interpreted as disabled or suspended when set to false.
            created : Filter by the Unix timestamp (in milliseconds) when this object was persisted.
            limit : Limit response to a specified positive number of objects. Ex. limit=10
        # /Connections/get_connections
        More info can be found here : https://www.adobe.io/apis/experienceplatform/home/api-reference.html
        """
        path = "/connections"
        params = {**kwargs}
        res = aepp._getData(self.endpoint+path,
                            headers=self.header, params=params)
        return res

    def getConnection(self, connection_id: str = None):
        """
        Retrieve a specific connection
        Arguments:
            connection_id : REQUIRED : ID of the connection to retrieve.
        """
        if connection_id is None:
            raise Exception("Expected a connection_id parameter")
        path = f"/connections/{connection_id}"
        res = aepp._getData(self.endpoint+path,
                            headers=self.header)
        return res

    def createConnection(self, data: dict = None):
        """
        create a new connection for data ingestion.
        """
        if type(data) != dict:
            raise TypeError("data should a dictionary")
        path = "/connections"
        res = aepp._postData(self.endpoint+path,
                             headers=self.header)
        return res

    def updateConnection(self, connection_id: str = None, data: dict = None):
        """
        Update the connection based on the data passed.
        Arguments:
            connection_id : REQUIRED : id of the connection to be updated
            data : REQUIRED : Object to update the data.
        Note : This is using the put statement.
        """
        if connection_id is None:
            raise Exception("Expected connection connection_id as argument")
        elif data is None or type(data) != dict:
            raise TypeError("Expected dictionary on the data argument")
        path = f"/connections/{connection_id}"
        res = aepp._putData(self.endpoint+path, data=data, headers=self.header)
        return res

    def deleteConnection(self, connection_id: str = None):
        """
        Delete a connection.
        Argument:
            connection_id : REQUIRED : ID of the connection to be deleted.
        """
        if connection_id is None:
            raise Exception("Expected connection connection_id as argument")
        path = f"/connections/{connection_id}"
        res = aepp._deleteData(self.endpoint+path, headers=self.header)
        return res

    def getConnectionDataSets(self, connection_id: str = None):
        """
        Return the dataSet attached to the conenction
        Arguments:
            connection_id : REQUIRED : ID of the connection.
        """
        if connection_id is None:
            raise Exception("Expected connection connection_id as argument")
        path = f"/connections/{connection_id}/dataSets"
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res

    def getDataSets(self, **kwargs):
        """
        Return a list of a datasets.
        Possible kwargs:
            state : The state related to a dataset.
            created : Filter by the Unix timestamp (in milliseconds) when this object was persisted.
            updated : Filter by the Unix timestamp (in milliseconds) for the time of last modification.
            name : Filter by the a descriptive, human-readable name for this DataSet.
            namespace : One of the registered platform acronyms that identify the platform.
            version : Filter by Semantic version of the account. Updated when the object is modified.
            property : Regex used to filter objects in the response. Ex. property=name~^test.
            # /Datasets/get_data_sets
            more possibilities : https://www.adobe.io/apis/experienceplatform/home/api-reference.html
        """
        path = "/dataSets"
        params = {**kwargs}
        res = aepp._getData(self.endpoint+path,
                            headers=self.header, params=params)
        try:
            self.data.table_names = {
                res[key]['name']: res[key]['tags']['adobe/pqs/table'] for key in res}
            self.data.schema_ref = {
                res[key]['name']: res[key]['schemaRef']
                for key in res if 'schemaRef' in res[key].keys()
            }
            self.data.ids = {
                res[key]['name']: key for key in res
            }
        except Exception as e:
            print(e)
            print("Couldn't populate the data object from the instance.")
        return res

    def createDataSets(self, data: dict = None, **kwargs):
        """
        Create a new dataSets.
        Arguments:
            data : REQUIRED : Data set to be posted
        possible kwargs
            requestDataSource : Set to true if you want Catalog to create a dataSource on your behalf; otherwise, pass a dataSourceId in the body.
        """
        path = "/dataSets"
        if data is None or isinstance(data, dict) == False:
            raise Exception("Excepting data of type dictionary")
        params = {"requestDataSource": kwargs.get("requestDataSource", False)}
        res = aepp._postData(self.endpoint+path, params=params,
                             data=data, headers=self.header)
        return res

    def getDataSet(self, dataset_id: str = None):
        """
        Return a single dataset.
        Arguments:
            dataset_id : REQUIRED : Id of the dataset to be retrieved.
        """
        if dataset_id is None:
            raise Exception("Expected a dataset_id argument")
        path = f"/dataSets/{dataset_id}"
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res

    def deleteDataSet(self, dataset_id: str = None):
        """
        Delete a dataset by its id.
        Arguments:
            dataset_id : REQUIRED : Id of the dataset to be deleted.
        """
        if dataset_id is None:
            raise Exception("Expected a dataset_id argument")
        path = f"/dataSets/{dataset_id}"
        res = aepp._deleteData(self.endpoint+path, headers=self.header)
        return res

    def getDataSetViews(self, dataset_id: str = None, **kwargs):
        """
        Get views of the datasets.
        Arguments:
            dataset_id : REQUIRED : Id of the dataset to be looked down.
        Possible kwargs:
            limit : Limit response to a specified positive number of objects. Ex. limit=10
            orderBy : Sort parameter and direction for sorting the response. Ex. orderBy=asc:created,updated.
            start : Returns results from a specific offset of objects. This was previously called offset. Ex. start=3.
            property : Regex used to filter objects in the response. Ex. property=name~^test.
        """
        if dataset_id is None:
            raise Exception("Expected a dataset_id argument")
        path = f"dataSets/{dataset_id}/views"
        params = {**kwargs}
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res

    def getDataSetView(self, dataset_id: str = None, view_id: str = None):
        """
        Get a specific view on a specific dataset.
        Arguments:
            dataset_id : REQUIRED : ID of the dataset to be looked down.
            view_id : REQUIRED : ID of the view to be look upon.
        """
        if dataset_id is None or view_id is None:
            raise Exception("Expected a dataset_id and an view_id argument")
        path = f"dataSets/{dataset_id}/views/{view_id}"
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res

    def getDataSetViewFiles(self, dataset_id: str = None, view_id: str = None):
        """
        Returns the list of files attached to a view in a Dataset.
        Arguments:
            dataset_id : REQUIRED : ID of the dataset to be looked down.
            view_id : REQUIRED : ID of the view to be look upon.
        """
        if dataset_id is None or view_id is None:
            raise Exception("Expected a dataset_id and an view_id argument")
        path = f"dataSets/{dataset_id}/views/{view_id}/files"
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res



# Internal Library
import aepp
from aepp import config
from pg import DB
from copy import deepcopy


class QueryService:
    """
    QueryService class is to be used 
    """

    def __init__(self, **kwargs):
        self.header = deepcopy(aepp.config._header)
        self.header.update({"Accept": "application/json"})
        self.header.update(**kwargs)
        self.endpoint = config._endpoint+config._endpoint_query

    def connection(self):
        """
        Create a connection for interactive interface. 
        """
        path = "/connection_parameters"
        res = aepp._getData(
            config._endpoint+config._endpoint_query+path, headers=self.header)
        return res

    def getQueries(self, orderby: str = None, limit: int = 1000, start: int = None, **kwargs):
        """
        Retrieve the queries from your organization.
        Arguments:
            orderby : OPTIONAL : possibility to order by "created" or "updated".
            prepend with "+" for ASC and "-" for DESC. To be escaped (default: -)
            limit : OPTIONAL : number of of records to fetch per page. (default 1000)
            start: OPTIONAL : when to start (depending on your orderby)

        possible kwargs:
            excludeSoftDeleted: Whether to include any queries that have been soft deleted. Defaults to true.
            excludeHidden : Whether to include any queries that have been found to be not interesting, as they were not user driven. Examples include CURSOR definitions, FETCH, and Metadata queries. Defaults to true.
            isPrevLink : Field indicating if the URI is a previous link.
        """
        path = '/queries'
        arguments = {'limit': 1000}
        if orderby is not None:
            if orderby == "+":
                orderby = "%2B"
            arguments['orderby'] = orderby
        if start is not None:
            arguments['start'] = start
        if limit is not None:
            arguments['limit'] = limit
        if len(kwargs.keys()) > 0:
            arguments['excludeSoftDeleted'] = kwargs.get(
                "excludeSoftDeleted", True)
            arguments['excludeHidden'] = kwargs.get(
                "excludeHidden", True)
            arguments['isPrevLink'] = kwargs.get(
                "isPrevLink", '')
        res = aepp._getData(
            config._endpoint+config._endpoint_query+path, params=arguments, headers=self.header)
        return res

    def postQueries(self, name: str = None, dbname: str = None, sql_query: str = None, templateId: str = None, queryParameters: dict = None, ctasParameters: dict = None, description: str = ""):
        """
        Create a query.
        Arguments:
            name : REQUIRED : Name of the query
            dbname : REQUIRED : the dataset name
            sql: REQUIRED : the SQL query as a string.
            queryParameters : OPTIONAL : in case you are using template, providing the paramter in a dictionary.
            ctasParameters: OPTIONAL : in case you want to create a dataset out of that query, dictionary is required with "datasetName" and "description".
        """
        path = '/queries'
        if sql_query is None or name is None:
            raise AttributeError("You are missing required arguments.")
        if type(ctasParameters) is dict:
            if "datasetName" not in ctasParameters.keys() or "description" not in ctasParameters.keys():
                raise KeyError(
                    'Expecting "datasetName" and "description" as part of the the ctasParameters dictionary.')
        data = {
            "name": name,
            "description": description,
            "dbName": dbname,
            "sql": sql_query,
        }
        if templateId is not None:
            data["templateId"] = templateId
        if queryParameters is not None:
            data["queryParameters"] = queryParameters
        if ctasParameters is not None:
            data["ctasParameters"] = ctasParameters
        res = aepp._postData(
            config._endpoint+config._endpoint_query+path, data=data, headers=self.header)
        return res

    def getQuery(self, queryid: str = None):
        """
        Request the query status by ID.
        Argument : 
            queryid : REQUIRED : the query id to check
        """
        if queryid is None:
            raise AttributeError('Expected "queryid" to be filled')
        path = '/queries/{queryid}'
        res = aepp._getData(
            config._endpoint+config._endpoint_query+path, headers=self.header)
        return res

    def getSchedules(self, **kwargs):
        """
        Retrieve a list of scheduled queries for the AEP instance.
        possibile kwargs:
            orderby : + for ASC and - for DESC
            limit : number of record to fetch
            start : specific start (use with orderby) 
            property : Comma-separated filters.
                created ; ex : created>2017-04-05T13:30:00Z
                templateId ; ex : templateId==123412354
                userId ; ex : userId==12341235
        """
        if kwargs.get("orderby", None) is not None:
            if kwargs.get("orderby", "-") == "+":
                kwargs.get("orderby", "-") == "%2B"
        path = "/schedules"
        params = {**kwargs}
        res = aepp._getData(self.endpoint + path, params=params)
        return res

    def getSchedule(self, scheduleId: str = None):
        """
        Get a details about a schedule query.
        Arguments:
            scheduleId : REQUIRED : the schedule id
        """
        if scheduleId is None:
            raise Exception("scheduleId is required")
        path = f"/schedules/{scheduleId}"
        res = aepp._getData(self.endpoint + path)
        return res

    def createSchedule(self, scheduleQuery: dict = None):
        """
        Create a scheduled query.
        Arguments:
            scheduleQuery: REQUIRED : a dictionary containing the query and the schedule. 
        """
        if type(scheduleQuery) != dict or scheduleQuery is None:
            raise Exception(
                "scheduleQuery is required and should be dictionary. ")
        if "query" not in scheduleQuery.keys() or "schedule" not in scheduleQuery.keys():
            raise Exception(
                'scheduleQuery dictonary is expected to contain "schedule" or "query" keys.')
        path = "/schedules"
        res = aepp._postData(self.endpoint + path, data=scheduleQuery)
        return res

    def deleteSchedule(self, scheduleId: str = None):
        """
        Delete a schedule query.
        Arguments: 
            scheduleId : REQUIRED : id of the schedule. 
        """
        path = f"/schedules/{scheduleId}"
        res = aepp._deleteData(self.endpoint + path)
        return res

    def getTemplates(self, **kwargs):
        """
        Retrieve the list of template for this instance.
        possible kwargs:
            limit : Hint on number of records to fetch per page. default (50)
            orderby : Field to order results by. Supported fields: created, updated. Prepend property name with + for ASC,- for DESC order. Default is -created.
            start : Start value of property specified using orderby.
            property : Comma-separated filters.List of properties that allow filtering:
                    name
                    userId
                    lastUpdatedBy
                operations:
                    ‘~’ (contains). This operator can only be used on the name property.
                    ‘==’ (equal to). This operator can be used on both the userId and the lastUpdatedBy properties.
        more details here : https://www.adobe.io/apis/experienceplatform/home/api-reference.html#/Query-Templates/get_query_templates
        """
        path = "/query-templates"
        params = {**kwargs}
        params['limit'] = kwargs.get('limit', 50)
        res = aepp._getData(self.endpoint + path,
                            headers=self.header, params=params)
        return res

    def createQueryTemplate(self, queryData: dict = None):
        """
        Create a query template based on the dictionary passed.
        Arguments: 
            queryData : REQUIED : An object that contains "sql", "queryParameter" and "name" keys.
        more info : https://www.adobe.io/apis/experienceplatform/home/api-reference.html#/Query-Templates/create_query_template
        """
        path = "/query-templates"
        if isinstance(queryData, dict):
            if "sql" not in queryData.keys() or "queryParameter" not in queryData.keys() or "name" not in queryData.keys():
                raise KeyError(
                    "Minimum key value are not respected.\nPlease see here for more info :\nhttps://www.adobe.io/apis/experienceplatform/home/api-reference.html#/Query-Templates/create_query_template ")
        else:
            raise Exception("expected a dictionary for queryData")
        res = aepp._postData(self.endpoint + path,
                             headers=self.header, data=queryData)
        return res


class InteractiveQuery:
    """
    Provide the instance connected to PostgreSQL database and you can return the result directly in your notebook.
    This class requires that you have used connection method in the QueryService.
    The object returned by the connection method should be used when creating this object.  

    """
    config_object = {}

    def __init__(self, conn_object: dict = None):
        if conn_object is None:
            raise AttributeError(
                "You are missing the conn_object. Use the QueryService to retrieve the object.")
        self.dbname = conn_object['dbName']
        self.host = conn_object['host']
        self.port = conn_object['port']
        self.user = conn_object['username']
        self.passwd = conn_object['token']
        self.websocketHost = conn_object['websocketHost']
        self.config_object = {
            "dbname": self.dbname,
            "host": self.host,
            "user": self.user,
            "passwd": self.passwd,
            "port": self.port,
        }
        self.db = DB(**self.config_object)

    def query(self, sql: str = None, output: str = "dataframe"):
        """
        Query the database and return different type of data, depending the format parameters.
        Requests are limited to return 50 K rows
        Arguments:
            sql : REQUIRED : the SQL request you want to realize. 
            output : OPTIONAL : the format you would like to be returned.
            Possible format:
                "raw" : return the instance of the query object.
                "dataframe" : return a dataframe with the data. (default)
        """
        if sql is None:
            raise Exception("Required a SQL query")
        query = self.db.query(sql)
        if output == "raw":
            return query
        elif output == "dataframe":
            data = query.getresult()
            columns = query.listfields()
            df = aepp.pd.DataFrame(data, columns=columns)
            return df
        else:
            raise KeyError("You didn't specify a correct value.")

    def transformToDataFrame(self, query: object = None):
        """
        This will return you a dataFrame 
        """
        data = query.getresult()
        columns = query.listfields()
        df = aepp.pd.DataFrame(data, columns=columns)
        return df



import aepp
_org_id, _api_key, _tech_id, _pathToKey, _secret, _companyid = "", "", "", "", "", ""
_TokenEndpoint = "https://ims-na1.adobelogin.com/ims/exchange/jwt"
_cwd = aepp.Path.as_posix(aepp.Path.cwd())
_date_limit = 0
_token = ''
_header = {"Accept": "application/json",
           "Content-Type": "application/json",
           "Authorization": "",
           "X-Api-Key": _api_key,
           "x-gw-ims-org-id": _org_id,
           "x-sandbox-name": "prod"
           }

# endpoints
_endpoint = "https://platform.adobe.io"
_endpoint_schema = "/data/foundation/schemaregistry"
_endpoint_query = "/data/foundation/query"
_endpoint_catalog = "/data/foundation/catalog"
_endpoint_dule = "/data/foundation/dulepolicy"
_endpoint_export = "/data/foundation/export"
_endpoint_segmentation = "/data/core/ups"



import aepp
from aepp import config
from copy import deepcopy
import typing


class Dule:
    """
    Class to manage and retrieve DULE policy.
    This is based on the following API reference : https://www.adobe.io/apis/experienceplatform/home/api-reference.html#/
    """

    def __init__(self, **kwargs):
        self.header = deepcopy(config._header)
        self.header.update(**kwargs)
        self.endpoint = config._endpoint+config._endpoint_dule

    def getPoliciesCore(self, **kwargs):
        """
        Returns the core policies in place in the Organization.
        Possible kwargs:
            limit : A positive integer, providing a hint as to the maximum number of resources to return in one page of results.
            property : Filter responses based on a property and optional existence or relational values. 
            Only the ‘name’ property is supported for core resources. 
            For custom resources, additional supported property values include 'status’, 'created’, 'createdClient’, 'createdUser’, 'updated’, 'updatedClient’, and 'updatedUser’
            orderby : A comma-separated list of properties by which the returned list of resources will be sorted.
            start : Requests items whose ‘orderby’ property value are strictly greater than the supplied ‘start’ value.
            duleLabels : A comma-separated list of DULE labels. Return only those policies whose "deny" expression references any of the labels in this list
            marketingAction : Restrict returned policies to those that reference the given marketing action.
        """
        path = "/policies/core"
        params = {**kwargs}
        res = aepp._getData(self.endpoint+path,
                            params=params, headers=self.header)
        return res

    def getPoliciesCoreId(self, policy_id: str = None):
        """
        Return a specific core policy by its id.
        Arguments:
            policy_id : REQUIRED : policy_id to retrieve.
        """
        if policy_id is None:
            raise Exception("Expected a policy id")
        path = f"/policies/core/{policy_id}"
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res

    def getPoliciesCustom(self, **kwargs):
        """
        Returns the custom policies in place in the Organization.
        Possible kwargs:
            limit : A positive integer, providing a hint as to the maximum number of resources to return in one page of results.
            property : Filter responses based on a property and optional existence or relational values. 
            Only the ‘name’ property is supported for core resources. 
            For custom resources, additional supported property values include 'status’, 'created’, 'createdClient’, 'createdUser’, 'updated’, 'updatedClient’, and 'updatedUser’
            orderby : A comma-separated list of properties by which the returned list of resources will be sorted.
            start : Requests items whose ‘orderby’ property value are strictly greater than the supplied ‘start’ value.
            duleLabels : A comma-separated list of DULE labels. Return only those policies whose "deny" expression references any of the labels in this list
            marketingAction : Restrict returned policies to those that reference the given marketing action.
        """
        path = "/policies/custom"
        params = {**kwargs}
        res = aepp._getData(self.endpoint+path,
                            params=params, headers=self.header)
        return res

    def getPoliciesCustom(self, policy_id: str = None):
        """
        Return a specific custom policy by its id.
        Arguments:
            policy_id: REQUIRED: policy_id to retrieve.
        """
        if policy_id is None:
            raise Exception("Expected a policy id")
        path = f"/policies/custom/{policy_id}"
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res

    def createPolicy(self, policy: typing.Union(dict, typing.IO) = None):
        """
        Create a custom policy.
        Arguments:
            policy : REQURED : A dictionary contaning the policy you would like to implement.
        """
        path = "/policies/custom"



# Internal Library
import aepp
from aepp import config
from copy import deepcopy
import typing
from concurrent import futures as _futures


class Segmentation:

    def __init__(self, **kwargs):
        self.header = deepcopy(aepp.config._header)
        self.header['Accept'] = "application/vnd.adobe.xdm+json"
        self.header.update(**kwargs)
        self.endpoint = config._endpoint+config._endpoint_segmentation

    def getSegments(self, **kwargs):
        """
        Return segment definitions in your experience platfom instance.
        Possible arguments:
            - limit : number of segment returned per page
        """
        params = {'limit': kwargs.get('limit', 100)}
        path = "/segment/definitions"
        res = aepp._getData(self.endpoint+path, headers=self.header)
        if 'segments' in res.keys():
            data = res['segments']
        else:
            data = []
        total_pages = res['page']['totalPages']
        if total_pages > 1:
            nb_request = total_pages
            max_workers = min((len(total_pages), 5))
            list_parameters = [{'page': str(
                x), **params} for x in range(2, total_pages+1)]
            urls = [self.endpoint+path for x in range(2, total_pages+1)]
            with _futures.ThreadPoolExecutor(max_workers) as executor:
                res = executor.map(lambda x, y: aepp._getData(
                    x, params=y), urls, list_parameters)
            res = list(res)
            append_data = [val for sublist in [data['data'] for data in res]
                           for val in sublist]  # flatten list of list
            data = data + append_data
        return data

    def getSegment(self, segment_id: str = None):
        """
        Return a specific segment definition.
        Argument:
            segment_id : REQUIRED : Segment ID of the segment to be retrieved.
        """
        if segment_id is None:
            raise Exception(
                "Expecting a segment ID to fetch the segment definition.")
        path = f"/segment/definitions/{segment_id}"
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res

    def createSegment(self, segment_data: dict = None):
        """
        Create a segment based on the information provided by the dictionary passed.
        Argument :
            segment_data : REQUIRED : Dictionary of the segment definition.
                require in the segment_data: name, description, expression, schema, ttlInDays
        """
        path = "/segment/definitions"
        if segment_data is None or type(segment_data) != dict:
            raise Exception(
                "Expecting data as dictionary format to update the segment.")
        checks = "name,description,expression,schema,ttlInDays".split(
            ',')  # mandatory elements in segment definition
        if len(set(checks) & set(segment_data.keys())) != len(checks):
            raise Exception(
                "Segment data doesn't hold one or several mandatory fields:\n\
                name, description, expression, schema, ttlInDays")
        res = aepp._postData(self.endpoint+path,
                             data=segment_data, headers=self.header)
        return res

    def deleteSegment(self, segment_id: str = None):
        """
        Delete a specific segment definition.
        Argument:
            segment_id : REQUIRED : Segment ID of the segment to be deleted.
        """
        if segment_id is None:
            raise Exception(
                "Expecting a segment ID to delete the segment.")
        path = f"/segment/definitions/{segment_id}"
        res = aepp._deleteData(self.endpoint+path, headers=self.header)
        return res

    def updateSegment(self, segment_id: str = None, segment_data: dict = None):
        """
        Update the segment characteristics from the definition pass to that method.
        Arguments:
            segment_id : REQUIRED : id of the segment to be udpated.
            segment_data : REQUIRED : Dictionary of the segment definition.
                require in the segment_data: name, description, expression, schema, ttlInDays
        """
        if segment_id is None:
            raise Exception(
                "Expecting a segment ID to update the segment.")
        elif segment_data is None or type(segment_data) != dict:
            raise Exception(
                "Expecting data as dictionary format to update the segment.")
        path = f"/segment/definitions/{segment_id}"
        checks = "name,description,expression,schema,ttlInDays".split(
            ',')  # mandatory elements in segment definition
        if len(set(checks) & set(segment_data.keys())) != len(checks):
            raise Exception(
                "Segment data doesn't hold one or several mandatory fields:\n\
                name, description, expression, schema, ttlInDays")
        update = aepp._postData(
            self.endpoint+path, headers=self.header, data=segment_data)
        return update

    def exportJobs(self, limit: int = 100, status: str = None):
        """
        Retrieve a list of all export jobs.
        Arguments:
            limit : OPTIONAL : number of jobs to be returned (default 100)
            status : OPTIONAL : status of the job (NEW, SUCCEEDED, FAILED)
        """
        path = "/export/jobs"
        params = {"limit": limit}
        if status is not None and status in ['NEW', 'SUCEEDED', 'FAILED']:
            params['status'] = status
        res = aepp._getData(self.endpoint+path,
                            params=params, headers=self.header)
        return res

    def createExport(self, export_data: dict = None):
        """
        Retrieve a list of all export jobs.
        Arguments:
            export_data : REQUIRED : number of jobs to be returned (default 100)
        """
        path = "/export/jobs"
        if export_data is None:
            raise Exception(
                "Expected export data to specify segment to export.")
        res = aepp._postData(self.endpoint+path,
                             data=export_data, headers=self.header)
        return res

    def getExport(self, export_id: str = None):
        """
        Retrieve a specific export Job.
        Arguments:
            export_id : REQUIRED : Export Job to be retrieved.
        """
        if export_id is None:
            raise Exception("Expected a export_id")
        path = f"/export/jobs/{export_id}"
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res

    def deleteExport(self, export_id: str = None):
        """
        Cancel or delete an export Job.
        Arguments:
            export_id : REQUIRED : Export Job to be deleted.
        """
        if export_id is None:
            raise Exception("Expected a export_id")
        path = f"/export/jobs/{export_id}"
        res = aepp._deleteData(self.endpoint+path, headers=self.header)
        return res

    def searchNamespaces(self, query: str = None, schema: str = "_xdm.context.segmentdefinition", **kwargs):
        """
        Return a list of search count results, queried across all namespaces.
        Arguments:
            query : REQUIRED : the search query.
            schema : OPTIONAL : The schema class value associated with the search objects. (default _xdm.context.segmentdefinition)
        """
        path = "/search/namespaces"
        if query is None:
            raise Exception("Expected a query to search for.")
        params = {"schema.name": schema, "s": query}
        self.header['x-ups-search-version'] = "1.0"
        res = aepp._getData(self.endpoint+path,
                            headers=self.header, params=params)
        del self.header["x-ups-search-version"]
        return res

    def searchEntity(self, query: str = None, namespace: str = "ECID", entityId: str = None, schema: str = "_xdm.context.segmentdefinition", **kwargs):
        """
        Return the list of objects that are contained  within a namespace.
        Arguments:
            query : REQUIRED : the search query.
            schema : OPTIONAL : The schema class value associated with the search objects.(defaul _xdm.context.segmentdefinition)
            namespace : OPTIONAL : The namespace you want to search within (default ECID)
            entityId : OPTIONAL : The ID of the folder you want to search for external segments in
        possible kwargs:
            limit : maximum number of result per page. Max 50.
            page : page to be retrieved (start at 0)
            page_limit : maximum number of pages retrieved.
        """
        path = "/search/entities"
        if query is None:
            raise Exception("Expected a query to search for.")
        limit = kwargs.get('limit', 50)
        page = kwargs.get("page", 0)
        page_limit = kwargs.get("page_limit", 0)
        self.header['x-ups-search-version'] = "1.0"
        params = {"schemaClass": schema,
                  "namespace": namespace, "s": query, "entityId": entityId, "limit": limit, "page": page}
        res = aepp._getData(self.endpoint+path,
                            headers=self.header, params=params)
        data = res['entities']
        curr_page = res['page']["pageOffset"]
        total_pages = res['page']["totalPages"]
        while curr_page <= page_limit - 1 or curr_page == total_pages:
            res = aepp._getData(self.endpoint+path,
                                headers=self.header, params=params)
            data += res['entities']
            curr_page = res['page']["pageOffset"]
            total_pages = res['page']["totalPages"]
        del self.header["x-ups-search-version"]
        return data

    def getSchedules(self):
        """
        Return the list of scheduled segments.
        """
        path = "/config/schedules"
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res

    def createSchedule(self, schedule_data: dict = None):
        """
        Schedule a segment to run.
        Arguments:
            schedule_data : REQUIRED : Definition of the schedule.
            Should contains name, type, properties, schedule.
        """
        path = "/config/schedules"
        if schedule_data is None or type(schedule_data) != dict:
            raise Exception(
                "Expected a dictionary data for setting the segment schedule.")
        min_requirements = "name,type,properties,schedule".split(',')
        if len(set(min_requirements) & set(schedule_data.keys())) != len(min_requirements):
            raise Exception(
                "Missing one minimal requirements : name, type, properties, schedule")
        res = aepp._postData(self.endpoint+path,
                             data=schedule_data, headers=self.header)
        return res

    def getSchedule(self, schedule_id: str = None):
        """
        Get a specific schedule definition.
        Argument:
            schedule_id : REQUIRED : Segment ID to be retrieved.
        """
        if schedule_id is None:
            raise Exception("Expected a schedule_id")
        path = f"/config/schedules/{schedule_id}"
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res

    def deleteSchedule(self, schedule_id: str = None):
        """
        Delete a specific schedule definition.
        Argument:
            schedule_id : REQUIRED : Segment ID to be deleted.
        """
        if schedule_id is None:
            raise Exception("Expected a schedule_id")
        path = f"/config/schedules/{schedule_id}"
        res = aepp._deleteData(self.endpoint+path, headers=self.header)
        return res

    def getJobs(self, name: str = None, status: str = None, limit: int = 100, **kwargs):
        """
        Returns the list of segment jobs.
        Arguments: 
            name : OPTIONAL : Name of the snapshot
            status : OPTIONAL : Status of the job (PROCESSING,)
            limit : OPTIONAL : Amount of jobs to be retrieved
        """
        path = "/segment/jobs"
        params = {"snapshot.name": name, "status": status, "limit": limit}
        res = aepp._getData(self.endpoint+path,
                            params=params, headers=self.header)
        return res

    def createJob(self, job_data: list = None):
        """
        Create a new job for a segment.
        Argument: 
            job_data : REQUIRED : a list describing the job to be created.
        """
        path = "/segment/jobs"
        if job_data is None or type(job_data) != list:
            raise Exception("Expecting a list of segment ID to run.")
        res = aepp._postData(self.endpoint+path,
                             data=job_data, headers=self.header)
        return res

    def getJob(self, job_id: str = None):
        """
        Retrieve a Segment job by ID.
        Argument:
            job_id: REQUIRED : The job ID to retrieve.
        """
        path = f"/segment/jobs/{job_id}"
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res

    def deleteJob(self, job_id: str = None):
        """
        deleteJob a Segment job by ID.
        Argument:
            job_id: REQUIRED : The job ID to delete.
        """
        path = f"/segment/jobs/{job_id}"
        res = aepp._deleteData(self.endpoint+path, headers=self.header)
        return res



import aepp
from aepp import config


def retrieveToken(verbose: bool = False, save: bool = False, **kwargs)->str:
    """ Retrieve the token by using the information provided by the user during the import importConfigFile function. 

    Argument : 
        verbose : OPTIONAL : Default False. If set to True, print information.
    """
    if config._pathToKey.startswith('/'):
        config._pathToKey = "."+config._pathToKey
    with open(aepp.Path(config._pathToKey), 'r') as f:
        private_key_unencrypted = f.read()
        header_jwt = {'cache-control': 'no-cache',
                      'content-type': 'application/x-www-form-urlencoded'}
    jwtPayload = {
        # Expiration set to 24 hours
        "exp": round(24*60*60 + int(aepp.time.time())),
        "iss": config._org_id,  # org_id
        "sub": config._tech_id,  # technical_account_id
        "https://ims-na1.adobelogin.com/s/ent_dataservices_sdk": True,
        "aud": "https://ims-na1.adobelogin.com/c/"+config._api_key
    }
    encoded_jwt = aepp.jwt.encode(
        jwtPayload, private_key_unencrypted, algorithm='RS256')  # working algorithm
    payload = {
        "client_id": config._api_key,
        "client_secret": config._secret,
        "jwt_token": encoded_jwt.decode("utf-8")
    }
    response = aepp.requests.post(config._TokenEndpoint,
                                  headers=header_jwt, data=payload)
    json_response = response.json()
    token = json_response['access_token']
    aepp.config._token = token
    aepp.config._header.update({"Authorization": "Bearer "+token})
    expire = json_response['expires_in']
    config._date_limit = aepp.time.time() + expire/1000 - \
        500  # end of time for the token
    if save:
        with open('token.txt', 'w') as f:  # save the token
            f.write(token)
    if verbose == True:
        print('token valid till : ' +
              aepp.time.ctime(aepp.time.time() + expire/1000))
        print('token has been saved here : ' +
              aepp.Path.as_posix(aepp.Path.cwd()))
    return token


def _checkToken(func):
    """    decorator that checks that the token is valid before calling the API    """
    def checking(*args, **kwargs):  # if function is not wrapped, will fire
        config._date_limit
        now = aepp.time.time()
        if now > config._date_limit - 1000:
            config._token = retrieveToken(*args, **kwargs)
            kwargs['headers']['Authorization'] = "Bearer "+config._token
            return func(*args, **kwargs)
        else:  # need to return the function for decorator to return something
            return func(*args, **kwargs)
    return checking  # return the function as object



import time
import json
from collections import defaultdict
from concurrent import futures
from copy import deepcopy
from typing import Union, IO
from datetime import datetime
# Non standard libraries
import pandas as pd
import requests
import jwt
from pathlib import Path
# Internal Library
from aepp import config
from aepp import adobeio_auth


__version__ = "0.0.3"

def createConfigFile(verbose: object = False)->None:
    """
    This function will create a 'config_admin.json' file where you can store your access data. 
    """
    json_data = {
        'org_id': '<orgID>',
        'api_key': "<APIkey>",
        'tech_id': "<something>@techacct.adobe.com",
        'secret': "<YourSecret>",
        'pathToKey': '<path/to/your/privatekey.key>',
    }
    with open('config_admin.json', 'w') as cf:
        cf.write(json.dumps(json_data, indent=4))
    if verbose:
        print(' file created at this location : ' +
              config._cwd + '/config_admin.json')


def importConfigFile(file: str)-> None:
    """
    This function will read the 'config_admin.json' to retrieve the information to be used by this module. 
    """
    with open(Path(file), 'r') as file:
        f = json.load(file)
        config._org_id = f['org_id']
        config._header["x-gw-ims-org-id"] = config._org_id
        config._api_key = f['api_key']
        config._header["X-Api-Key"] = config._api_key
        config._tech_id = f['tech_id']
        config._secret = f['secret']
        config._pathToKey = f['pathToKey']


@adobeio_auth._checkToken
def _getData(endpoint: str, params: dict = None, data: dict = None, headers: dict = None, *args, **kwargs):
    """
    Abstraction for getting data
    """
    if headers is None:
        headers = config._header
    if params == None and data == None:
        res = requests.get(endpoint, headers=headers)
    elif params != None and data == None:
        res = requests.get(endpoint, headers=headers, params=params)
    elif params == None and data != None:
        res = requests.get(endpoint, headers=headers, data=data)
    elif params != None and data != None:
        res = requests.get(endpoint, headers=headers,
                           params=params, data=data)
    try:
        res_json = res.json()
    except:
        res_json = {'error': 'Request Error'}
    return res_json


@adobeio_auth._checkToken
def _postData(endpoint: str, params: dict = None, data: dict = None, headers: dict = None, * args, **kwargs):
    """
    Abstraction for posting data
    """
    if headers is None:
        headers = config._header
    if params == None and data == None:
        res = requests.post(endpoint, headers=headers)
    elif params != None and data == None:
        res = requests.post(endpoint, headers=headers, params=params)
    elif params == None and data != None:
        res = requests.post(endpoint, headers=headers,
                            data=json.dumps(data))
    elif params != None and data != None:
        res = requests.post(endpoint, headers=headers,
                            params=params, data=json.dumps(data))
    try:
        res_json = res.json()
    except:
        res_json = {'error': 'Request Error'}
    return res_json


@adobeio_auth._checkToken
def _patchData(endpoint: str, params: dict = None, data: dict = None, headers: dict = None, *args, **kwargs):
    """
    Abstraction for patching data
    """
    if headers is None:
        headers = config._header
    if params == None and data == None:
        res = requests.patch(endpoint, headers=headers)
    elif params != None and data == None:
        res = requests.patch(endpoint, headers=headers, params=params)
    elif params == None and data != None:
        res = requests.patch(endpoint, headers=headers,
                             data=json.dumps(data))
    elif params != None and data != None:
        res = requests.patch(endpoint, headers=headers,
                             params=params, data=json.dumps(data))
    try:
        res_json = res.json()
    except:
        res_json = {'error': 'Request Error'}
    return res_json


@adobeio_auth._checkToken
def _deleteData(endpoint: str, params: dict = None, data=None, headers: dict = None, *args, **kwargs):
    """
    Abstraction for deleting data
    """
    if headers is None:
        headers = config._header
    if params == None and data == None:
        res = requests.delete(endpoint, headers=headers)
    elif params != None and data == None:
        res = requests.delete(endpoint, headers=headers, params=params)
    elif params == None and data != None:
        res = requests.delete(endpoint, headers=headers,
                              data=json.dumps(data))
    elif params != None and data != None:
        res = requests.delete(endpoint, headers=headers,
                              params=params, data=json.dumps(data=data))
    try:
        status_code = res.status_code
    except:
        status_code = {'error': 'Request Error'}
    return status_code


@adobeio_auth._checkToken
def _putData(endpoint: str, params: dict = None, data=None, headers: dict = None, *args, **kwargs):
    """
    Abstraction for deleting data
    """
    if headers is None:
        headers = config._header
    if params != None and data == None:
        res = requests.put(endpoint, headers=headers, params=params)
    elif params == None and data != None:
        res = requests.put(endpoint, headers=headers,
                           data=json.dumps(data))
    elif params != None and data != None:
        res = requests.put(endpoint, headers=headers,
                           params=params, data=json.dumps(data=data))
    try:
        status_code = res.json()
    except:
        status_code = {'error': 'Request Error'}
    return status_code


def home(product: str = None, limit: int = 50):
    """
    Return the IMS Organization setup and the container existing for the organization
    Arguments:
        product : OPTIONAL : specify one or more product contexts for which to return containers. If absent, containers for all contexts that you have rights to will be returned. The product parameter can be repeated for multiple contexts. An example of this parameter is product=acp
        limit : OPTIONAL : Optional limit on number of results returned (default = 50).
    """
    endpoint = config._endpoint+"/data/core/xcore/"
    params = {"product": product, "limit": limit}
    myHeader = deepcopy(config._header)
    myHeader["Accept"] = "application/vnd.adobe.platform.xcore.home.hal+json"
    res = _getData(endpoint, params=params, headers=myHeader)
    return res



# Internal Library
import aepp
from aepp import config
from copy import deepcopy
import re
import typing
from dataclasses import dataclass

json_extend = [{'op': 'replace',
                'path': '/meta:intendedToExtend',
                'value': ['https://ns.adobe.com/xdm/context/profile',
                          'https://ns.adobe.com/xdm/context/experienceevent']}]


@dataclass
class _Data:

    def __init__(self):
        self.ids = {}
        self.schemas = {}
        self.paths = {}


class Schema:
    """
    This class is a wrapper around the schema registry API for Adobe Experience Platform.
    More documentation on these endpoints can be found here : https://www.adobe.io/apis/experienceplatform/home/api-reference.html#!acpdr/swagger-specs/schema-registry.yaml

    """
    schemas = {}
    schemasPaths = {}
    _schemaClasses = {
        "event": "https://ns.adobe.com/xdm/context/experienceevent",
        "profile": "https://ns.adobe.com/xdm/context/profile"
    }

    def __init__(self, **kwargs):
        self.header = deepcopy(aepp.config._header)
        self.header['Accept'] = "application/vnd.adobe.xdm+json"
        self.header.update(**kwargs)
        self.endpoint = config._endpoint+config._endpoint_schema
        self.data = _Data()

    def _getPaths(self, myDict: dict, path: list = None, key: str = None, list_path: list = None, verbose: bool = False):
        """
        Function to find the different paths existing in a schema
        Arguments:
            myDict : REQUIRED : Schema that needs to be analyzed
            paths : Not required - path being built on run time
            key : key pass to the lower iteration in case.
            list_path : what is returned at the end of the recursion
            verbose : OPTIONAL : In case you need to see something
        """
        if path is None:
            path = []
        if list_path is None:
            list_path = []
        if type(myDict) == dict:
            if "properties" in myDict.keys():  # have the properties in dict
                if key is not None:
                    path.append(key)
                res = self._getPaths(myDict['properties'],
                                     key=key, path=path, list_path=list_path, verbose=verbose)
            elif type(myDict) == dict:  # 2 have not the property option
                if myDict.get("type", 'object') == "object":
                    for key, value in myDict.items():
                        recurs = self._getPaths(value, path=path,
                                                key=key, list_path=list_path, verbose=verbose)
                        if key in path:
                            path.pop()
                else:  # last level Not object type.
                    path.append(key)
                    full_path = ".".join(path)
                    if verbose:
                        print(full_path)
                    list_path.append(full_path)
                    path.pop()
                    return path
            return list_path

    def getStats(self):
        """
        Returns a list of the last actions realized on the Schema for this instance of AEP. 
        """
        path = '/stats/'
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res

    def getSchemas(self):
        """
        Returns a lit of the schemas hosted in this instance.
        """
        path = '/tenant/schemas/'
        res = aepp._getData(self.endpoint+path, headers=self.header)
        data = res['results']
        self.data.ids = {key['title']: key['$id'] for key in data}
        return data

    def getSchema(self, schema_id: str = None, version: int = 1, save: bool = False, findPaths: bool = False, **kwargs):
        """
        Get the Schema. Requires a schema id.
        Response provided depends on the header set, you can change the Accept header with kwargs.
        Arguments:
            schema_id : REQUIRED : $id or meta:altId
            version : OPTIONAL : Version of the Schema asked (default 1)
            save : OPTIONAL : save the result in json file (default False)
            findPaths : find the paths present in your schema.
        Possible kwargs:
            Accept : Accept header to change the type of response.
            # /Schemas/lookup_schema
            more details held here : https://www.adobe.io/apis/experienceplatform/home/api-reference.html
        """
        self.header["Accept"] = "application/vnd.adobe.xed-full+json; version=" + \
            str(version)
        if kwargs.get('Accept', None) is not None:
            header['Accept'] = kwargs.get('Accept', header['Accept'])
        self.header['Accept-Encoding'] = 'identity'
        if schema_id.startswith('https://'):
            from urllib import parse
            schema_id = parse.quote_plus(schema_id)
        path = f'/tenant/schemas/{schema_id}'
        res = aepp._getData(self.endpoint + path, headers=self.header)
        del self.header['Accept-Encoding']
        self.header['Accept'] = "application/json"
        if "title" not in res.keys():
            print('Issue with the request. See response.')
            return res
        if save:
            with open(f'{res["title"]}.json', 'w') as f:
                f.write(aepp.json.dumps(res, indent=4))
        self.data.schemas[res['title']] = res
        if findPaths:
            paths = self._getPaths(res)
            self.data.paths[res['title']] = paths
        return res

    def patchSchema(self, schema_id: str = None, changes: list = None, **kwargs):
        """
        Enable to patch the Schema with operation.
        Arguments:
            schema_id : REQUIRED : $id or meta:altId
            change : REQUIRED : List of changes that need to take place.
        information : http://jsonpatch.com/
        """
        path = f'/ tenant/schemas/{schema_id}'
        if schema_id is None:
            raise Exception("Require an ID for the schema")
        if type(changes) == dict:
            changes = list(changes)
        if schema_id.startswith('https://'):
            from urllib import parse
            schema_id = parse.quote_plus(schema_id)
        res = aepp._patchData(self.endpoint+path,
                              data=changes, headers=self.header)
        return res

    def putSchema(self, schema_id: str = None, changes: dict = None, **kwargs):
        """
        A PUT request essentially re-writes the schema, therefore the request body must include all fields required to create (POST) a schema.
        This is especially useful when updating a lot of information in the schema at once.
        Arguments:
            schema_id : REQUIRED : $id or meta:altId
            change : REQUIRED : dictionary of the new schema.
            It requires a allOf list that contains all the attributes that are required for creating a schema.
            #/Schemas/replace_schema
            More information on : https://www.adobe.io/apis/experienceplatform/home/api-reference.html
        """
        path = f'/ tenant/schemas/{schema_id}'
        if schema_id is None:
            raise Exception("Require an ID for the schema")
        if schema_id.startswith('https://'):
            from urllib import parse
            schema_id = parse.quote_plus(schema_id)
        res = aepp._putData(self.endpoint+path,
                            data=changes, headers=self.header)
        return res

    def deleteSchema(self, schema_id: str = None, **kwargs):
        """
        Delete the request
        Arguments:
            schema_id : REQUIRED : $id or meta:altId
            It requires a allOf list that contains all the attributes that are required for creating a schema.
            #/Schemas/replace_schema
            More information on : https://www.adobe.io/apis/experienceplatform/home/api-reference.html
        """
        path = f'/ tenant/schemas/{schema_id}'
        if schema_id is None:
            raise Exception("Require an ID for the schema")
        if schema_id.startswith('https://'):
            from urllib import parse
            schema_id = parse.quote_plus(schema_id)
        res = aepp._deleteData(self.endpoint+path, headers=self.header)
        return res

    def createSchema(self, schema: dict = None):
        """
        Create a Schema based on the data that are passed in the Argument.
        Arguments:
            schema : REQUIRED : The schema definition that needs to be created.
        """
        path = '/ tenant/schemas/'
        if type(schema) != dict:
            raise TypeError("Expecting a dictionary")
        if "allOf" not in schema.keys():
            raise Exception(
                "The schema must include an ‘allOf’ attribute (a list) referencing the $id of the base class the schema will implement.")
        res = aepp._postData(self.endpoint+path,
                             headers=self.header, data=schema)
        return res

    def findPath(self, schema: typing.Union[str, dict] = None, filter: str = None):
        """
        Return a list of the paths containing your filter.
        Arguments:
            schema : Schema JSON or title reference
            filter : REQUIRED : element you would like to find in your path.
            paths : OPTIONAL : list of all of the paths.
        """
        if type(schema) == str:
            if schema in self.data.paths.keys():
                paths = self.data.paths[schema]
            elif schema in self.data.schemas.keys():
                schema = self.data.schemas[schema]
                paths = self._getPaths(schema)
            else:
                raise Exception(
                    "Schema reference doesn't exist.\nMake sure to retrieve the schema you want to use beforehand")
        elif type(schema) == dict:
            try:
                paths = self._getPaths(schema)
                data.paths[schema['title']] = paths
            except:
                print('Issue with the schema reading. Verifiy your schema structure.')
                return None
        if filter is None:
            return paths
        elements = (element for element in paths if re.search(
            f".*{filter}.*", str(element)) is not None)
        return list(elements)

    def pathType(self, path: str = None, schema: typing.Union[dict, str] = None):
        """
        Returns the XDM Type of the path that has been passed with the schema.
        Arguments:
            path : REQUIRED : the path that you want to check. Ex : "_tenant.schema.something"
            schema : REQUIRED : Either the full schema definition or title reference to the schema.
            if the title reference has been provided, make sure you have downloaded it before with getSchema method.
        """
        if type(schema) == str:
            if schema not in self.data.schemas.keys():
                raise Exception(
                    "Expecting a reference to a schema already downloaded.")
            else:
                schema = self.data.schemas[schema]
        split_path = path.split('.')
        if len(path) == 0:
            return schema["meta:xdmType"]
        elif split_path[0] in schema.keys():
            result = self.pathType(
                '.'.join(split_path[1:]), schema[split_path[0]])
        elif "properties" in schema.keys():
            result = self.pathType('.'.join(split_path), schema['properties'])
        return result

    def getClasses(self):
        """
        return the classes of the AEP Instances.
        """
        path = '/tenant/classes/'
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res

    def getClass(self, class_id: str = None, version: int = 1):
        """
        Return a specific class.
        Arguments: 
            class_id : REQUIRED : the meta:altId or $id from the class
            version : OPTIONAL : the version of the class to retrieve.
        """
        if class_id is None:
            raise Exception("Require a class_id")
        header = self.header
        if class_id.startswith('https://'):
            from urllib import parse
            class_id = parse.quote_plus(class_id)
        self.header['Accept-Encoding'] = 'identity'
        self.header.update({
            "Accept": "application/vnd.adobe.xed-full+json; version="+str(version)})
        path = f'/tenant/classes/{class_id}'
        res = aepp._getData(self.endpoint + path, headers=header)
        del header['Accept-Encoding']
        self.header.update({
            "Accept": "application/json"})
        return res

    def getMixins(self):
        """
        returns the mixin of the account
        """
        path = '/tenant/mixins/'
        res = aepp._getData(self.endpoint+path, headers=self.header)
        return res

    def getMixin(self, mixin_id: str = None, version: int = 1):
        """
        Returns a specific mixin.
        Arguments:
            mixin_id : meta:altId or $id
            version : version of the mixin
        """
        header = self.header
        if mixin_id.startswith('https://'):
            from urllib import parse
            mixin_id = parse.quote_plus(mixin_id)
        self.header['Accept-Encoding'] = 'identity'
        self.header.update({
            "Accept": "application/vnd.adobe.xed-full+json; version="+str(version)})
        path = f'/tenant/mixins/{mixin_id}'
        res = aepp._getData(self.endpoint + path, headers=header)
        del header['Accept-Encoding']
        self.header.update({
            "Accept": "application/json"})
        return res

    def deleteMixin(self, mixin_id: str = None):
        """
        Arguments:
            schema_id : meta:altId or $id
        """
        if mixin_id is None:
            raise Exception("Require an ID")
        path = f'/tenant/mixins/{mixin_id}'
        if mixin_id.startswith('https://'):
            from urllib import parse
            mixin_id = parse.quote_plus(mixin_id)
        res = aepp._deleteData(self.endpoint+path, headers=self.header)
        return res

    def updateMixin(self, mixin_id: str = None, changes: list = None):
        """
        Update the mixin with the operation described in the changes.
        Arguments:
            schema_id : REQUIRED : meta:altId or $id
            changes : REQUIRED : dictionary on what to update on that mixin.
        """
        if mixin_id is None or changes is None:
            raise Exception("Require an ID and changes")
        path = f'/tenant/mixins/{mixin_id}'
        if type(changes) == dict:
            changes = list(changes)
        res = aepp._patchData(self.endpoint+path,
                              data=changes, headers=self.header)
        return res

    def getUnions(self, **kwargs):
        """
        Get all of the unions that has been set for the tenant.
        Returns a dictionary.

        Possibility to add option using kwargs
        """
        path = '/tenant/unions'
        params = {}
        if len(kwargs) > 0:
            for key in kwargs.key():
                if key == 'limit':
                    if int(kwargs['limit']) > 500:
                        kwargs['limit'] = 500
                params[key] = kwargs.get(key, '')
        res = aepp._getData(self.endpoint+path,
                            params=params, headers=self.header)
        data = res['results']  # issue when requesting directly results.
        return data

    def getUnion(self, union_id: str = None, version: int = 1):
        """
        Get a specific union type. Returns a dictionnary
        Arguments :
            union_id : REQUIRED :  meta:altId or $id
            version : OPTIONAL : version of the union schema required.
        """
        if union_id is None:
            raise Exception("Require an ID")
        if union_id.startswith('https://'):
            from urllib import parse
            union_id = parse.quote_plus(union_id)
        path = f'/tenant/unions/{union_id}'
        self.header.update({
            "Accept": "application/vnd.adobe.xed-full+json; version="+str(version)})
        res = aepp._getData(self.endpoint + path, headers=self.header)
        self.header.update({
            "Accept": "application/json"})
        return res

    def getXDMprofileSchema(self):
        """
        Returns a list of all schemas that are part of the XDM Individual Profile.
        """
        path = "/tenant/schemas?property=meta:immutableTags==union&property=meta:class==https://ns.adobe.com/xdm/context/profile"
        res = aepp._getData(self.endpoint + path, headers=self.header)
        return res





import time as _time
import json as _json
from collections import defaultdict as _defaultdict
from concurrent import futures as _futures
from copy import deepcopy as _deepcopy
from typing import Union, IO
from datetime import datetime
# Non standard libraries
import pandas as _pd
import requests as _requests
import jwt as _jwt
from pathlib import Path


# Set up default values
_org_id, _api_key, _tech_id, _pathToKey, _secret, _companyid = "", "", "", "", "", ""
_TokenEndpoint = "https://ims-na1.adobelogin.com/ims/exchange/jwt"
_cwd = Path.as_posix(Path.cwd())
_date_limit = 0
_token = ''
_header = {}


def createConfigFile(verbose: object = False)->None:
    """
    This function will create a 'config_admin.json' file where you can store your access data. 
    """
    json_data = {
        'org_id': '<orgID>',
        'api_key': "<APIkey>",
        'tech_id': "<something>@techacct.adobe.com",
        'secret': "<YourSecret>",
        'pathToKey': '<path/to/your/privatekey.key>',
    }
    with open('config_admin.json', 'w') as cf:
        cf.write(_json.dumps(json_data, indent=4))
    if verbose:
        print(' file created at this location : '+_cwd + '/config_admin.json')


def importConfigFile(file: str)-> None:
    """
    This function will read the 'config_admin.json' to retrieve the information to be used by this module. 
    """
    global _org_id
    global _api_key
    global _tech_id
    global _pathToKey
    global _secret
    global _endpoint
    with open(file, 'r') as file:
        f = _json.load(file)
        _org_id = f['org_id']
        _api_key = f['api_key']
        _tech_id = f['tech_id']
        _secret = f['secret']
        _pathToKey = f['pathToKey']


# Launch API Endpoint
_endpoint = 'https://platform.adobe.io'


def retrieveToken(verbose: bool = False, save: bool = False, **kwargs)->str:
    """ Retrieve the token by using the information provided by the user during the import importConfigFile function. 

    Argument : 
        verbose : OPTIONAL : Default False. If set to True, print information.
    """
    global _token
    with open(_pathToKey, 'r') as f:
        private_key_unencrypted = f.read()
        header_jwt = {'cache-control': 'no-cache',
                      'content-type': 'application/x-www-form-urlencoded'}
    jwtPayload = {
        # Expiration set to 24 hours
        "exp": round(24*60*60 + int(_time.time())),
        "iss": _org_id,  # org_id
        "sub": _tech_id,  # technical_account_id
        "https://ims-na1.adobelogin.com/s/ent_dataservices_sdk": True,
        "aud": "https://ims-na1.adobelogin.com/c/"+_api_key
    }
    encoded_jwt = _jwt.encode(
        jwtPayload, private_key_unencrypted, algorithm='RS256')  # working algorithm
    payload = {
        "client_id": _api_key,
        "client_secret": _secret,
        "jwt_token": encoded_jwt.decode("utf-8")
    }
    response = _requests.post(_TokenEndpoint, headers=header_jwt, data=payload)
    json_response = response.json()
    token = json_response['access_token']
    updateHeader(token=token)
    expire = json_response['expires_in']
    global _date_limit  # getting the scope right
    _date_limit = _time.time() + expire/1000 - 500  # end of time for the token
    if save:
        with open('token.txt', 'w') as f:  # save the token
            f.write(token)
    if verbose == True:
        print('token valid till : ' + _time.ctime(_time.time() + expire/1000))
        print('token has been saved here : ' + Path.as_posix(Path.cwd()))
    return token


def _checkToken(func):
    """    decorator that checks that the token is valid before calling the API    """
    def checking(*args, **kwargs):  # if function is not wrapped, will fire
        global _date_limit
        now = _time.time()
        if now > _date_limit - 1000:
            global _token
            _token = retrieveToken(*args, **kwargs)
            return func(*args, **kwargs)
        else:  # need to return the function for decorator to return something
            return func(*args, **kwargs)
    return checking  # return the function as object

###


@_checkToken
def _getData(endpoint: str, params: dict = None, data=None, header_info=None, *args, **kwargs):
    """
    Abstraction for getting data
    """
    global _header
    if header_info is not None:
        _header['Accept'] = _header['Accept']+header_info
    if params == None and data == None:
        res = _requests.get(endpoint, headers=_header)
    elif params != None and data == None:
        res = _requests.get(endpoint, headers=_header, params=params)
    elif params == None and data != None:
        res = _requests.get(endpoint, headers=_header, data=data)
    elif params != None and data != None:
        res = _requests.get(endpoint, headers=_header,
                            params=params, data=data)
    try:
        json = res.json()
    except:
        json = {'error': ['Request Error']}
    return json


@_checkToken
def _postData(endpoint: str, params: dict = None, data=None, *args, **kwargs):
    """
    Abstraction for getting data
    """
    global _header
    if params == None and data == None:
        res = _requests.post(endpoint, headers=_header)
    elif params != None and data == None:
        res = _requests.post(endpoint, headers=_header, params=params)
    elif params == None and data != None:
        res = _requests.post(endpoint, headers=_header, data=_json.dumps(data))
    elif params != None and data != None:
        res = _requests.post(endpoint, headers=_header,
                             params=params, data=_json.dumps(data=data))
    try:
        json = res.json()
    except:
        json = {'error': ['Request Error']}
    return json


@_checkToken
def _patchData(endpoint: str, params: dict = None, data=None, *args, **kwargs):
    """
    Abstraction for getting data
    """
    global _header
    if params == None and data == None:
        res = _requests.patch(endpoint, headers=_header)
    elif params != None and data == None:
        res = _requests.patch(endpoint, headers=_header, params=params)
    elif params == None and data != None:
        res = _requests.patch(endpoint, headers=_header,
                              data=_json.dumps(data))
    elif params != None and data != None:
        res = _requests.patch(endpoint, headers=_header,
                              params=params, data=_json.dumps(data=data))
    try:
        json = res.json()
    except:
        json = {'error': ['Request Error']}
    return json


@_checkToken
def _deleteData(endpoint: str, params: dict = None, data=None, *args, **kwargs):
    """
    Abstraction for getting data
    """
    global _header
    if params == None and data == None:
        res = _requests.delete(endpoint, headers=_header)
    elif params != None and data == None:
        res = _requests.delete(endpoint, headers=_header, params=params)
    elif params == None and data != None:
        res = _requests.delete(endpoint, headers=_header,
                               data=_json.dumps(data))
    elif params != None and data != None:
        res = _requests.delete(endpoint, headers=_header,
                               params=params, data=_json.dumps(data=data))
    try:
        status_code = res.status_code
    except:
        status_code = {'error': ['Request Error']}
    return status_code


def updateHeader(companyid: str = None, token: str = _token, **kwargs)->None:
    """ update the header when new token is generated
    This would be mandatory id you retrieved the company ID with the option "all". 
    Retrieving the company ID with option first or with a given position 
    will automatically call this method. 
    """
    global _header
    global _api_key
    global _companyid
    global _token
    global _org_id
    if token:
        _token = token
    _header = {"Accept": "application/vnd.adobe.xed-id+json",
               "Content-Type": "application/json",
               "Authorization": "Bearer "+_token,
               "X-Api-Key": _api_key,
               "x-gw-ims-org-id": _org_id
               }


_endpoint_schema = '/data/foundation/schemaregistry'


def getStats():
    path = '/stats/'
    res = _getData(_endpoint+_endpoint_schema+path)
    return res


def getSchemas():
    path = '/tenant/schemas/'
    res = _getData(_endpoint+_endpoint_schema+path)
    return res


def deleteSchema(schema_id: str = None):
    """
    Arguments:
        schema_id : meta:altId returned
    """
    if schema_id is None:
        raise Exception("Require an ID")
    path = f'/tenant/schemas/{schema_id}'
    res = _deleteData(_endpoint+_endpoint_schema+path)
    return res


def getMixins():
    path = '/tenant/mixins/'
    res = _getData(_endpoint+_endpoint_schema+path)
    return res


def getMixin(mixin_id: str = None, version: str = '1'):
    """
    Arguments:
        mixin_id : meta:altId
        version : version of the mixin
    """
    header_info = ';version='+str(version)
    path = f'/tenant/mixins/{mixin_id}'
    res = _getData(_endpoint+_endpoint_schema+path, header_info=header_info)
    return res


def deleteMixin(mixin_id: str = None):
    """
    Arguments:
        schema_id : meta:altId returned
    """
    if mixin_id is None:
        raise Exception("Require an ID")
    path = f'/tenant/mixins/{mixin_id}'
    res = _deleteData(_endpoint+_endpoint_schema+path)
    return res


def updateMixin(mixin_id: str = None, data: dict = None):
    """
    Arguments:
        schema_id : REQUIRED : meta:altId returned
        data : REQUIRED : dictionary on what to update on that mixin. 
    """
    if mixin_id is None or data is None:
        raise Exception("Require an ID and data")
    path = f'/tenant/mixins/{mixin_id}'
    res = _patchData(_endpoint+_endpoint_schema+path, data=data)
    return res


json_extend = [{'op': 'replace',
                'path': '/meta:intendedToExtend',
                'value': ['https://ns.adobe.com/xdm/context/profile',
                          'https://ns.adobe.com/xdm/context/experienceevent']}]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aepp",  # Replace with your own username
    version="0.0.3",
    author="Julien Piccini",
    author_email="piccini.julien@gmail.com",
    description="Package to manage AEP API endpoint and some helper functions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pitchmuc/aep",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Topic :: Utilities",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries",
        "Development Status :: 2 - Pre-Alpha"
    ],
    python_requires='>=3.6',
)
